/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/*
  Draw things with a sketchy, hand-drawn feel.

  By Carl Worth <cworth@redhat.com> following an original concept
  by Diana Fong <dfong@redhat.com>.
*/

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <glib.h>

#include "caligula-draw-sketchy.h"
#include "caligula-rand.h"

typedef struct pt {
    double x;
    double y;
} pt_t;

#define MAX_PTS		100
typedef struct poly {
    pt_t pts[MAX_PTS];
    int num_pts;
} poly_t;

#define MAX_POLYS	10
typedef struct poly_list {
    poly_t polys[MAX_POLYS];
    int num_polys;
} poly_list_t;

typedef struct bend_context {
    cairo_t *cr;

    int first;
    int closing;

    pt_t pt_first;
    pt_t pt_prev;
} bend_context_t;

/* Return a random positive double within 0 <= value < max */
static double
rand_double (double max)
{
    return max * caligula_rand() / (CALIGULA_RAND_MAX + 1.0);
}

/* Return the bearing of the line segment from a to b */
static double
direction (pt_t a, pt_t b)
{
    return atan2 (b.y - a.y, b.x - a.x);
}

static double
distance (pt_t a, pt_t b)
{
    double dx = a.x - b.x;
    double dy = a.y - b.y;
    return sqrt (dx*dx + dy*dy);
}

static void
pt_from_polar (pt_t *pt, double r, double theta)
{
    pt->x = r * cos (theta);
    pt->y = r * sin (theta);
}

static void
pt_tween (pt_t *res, pt_t a, pt_t b, double ratio)
{
    res->x = a.x + ratio * (b.x - a.x);
    res->y = a.y + ratio * (b.y - a.y);
}

/* Move a point so that it lies somewhere within the wedge centered at
 * the original point and defined by the given radius, bearing, and
 * angular width.
 */
static void
nudge (pt_t *pt, double radius,
       double angle_center, double angle_width)
{
    double r, theta;
    pt_t off;

    r = rand_double (radius);
    theta = angle_center + rand_double (angle_width) - angle_width / 2.0;

    pt_from_polar (&off, r, theta);

    pt->x += off.x;
    pt->y += off.y;
}

static void
bend_move_to (bend_context_t *context,
	      double          x,
	      double          y)
{
    context->pt_first.x = x;
    context->pt_first.y = y;

    context->pt_prev.x = x;
    context->pt_prev.y = y;

    if (context->first) {
	cairo_new_path (context->cr);
	context->first = 0;
    }
}

static void
bend_line_to (bend_context_t *context,
	      double          x,
	      double          y)
{
    pt_t a, b, c, d;
    double length;

    if (context->first)
	return bend_move_to (context, x, y);

    a = context->pt_prev;
    d.x = x;
    d.y = y;

    pt_tween (&b, a, d, 1.0/3.0);
    pt_tween (&c, a, d, 2.0/3.0);

    length = distance (a, b);
    {
	nudge (&a, 5, direction(b, a), M_PI / 10.);

	nudge (&b, 10, 0, 2*M_PI);
	nudge (&c, 10, 0, 2*M_PI);

/*
	if (! context->closing)
	    nudge (&d, 1./5. * length, direction(a, b), M_PI / 10.);
*/
    }

    cairo_move_to (context->cr,
		   a.x, a.y);
    cairo_curve_to (context->cr,
		    b.x, b.y,
		    c.x, c.y,
		    d.x, d.y);

    context->pt_prev.x = x;
    context->pt_prev.y = y;
}

static void
bend_close_path (bend_context_t *context)
{
    context->closing = 1;

    bend_line_to (context, context->pt_first.x, context->pt_first.y);
}

static void
bend_path (cairo_t *cr)
{
    bend_context_t context;
    cairo_path_t *path;
    int i;

    context.cr = cr;
    context.first = 1;
    context.closing = 0;
    
    path = cairo_copy_path_flat (cr);
    for (i=0; i < path->num_data; i += path->data[i].header.length) {
	cairo_path_data_t *data = &path->data[i];

	switch (data->header.type) {
	case CAIRO_PATH_MOVE_TO:
	    bend_move_to (&context,
			  data[1].point.x, data[1].point.y);
	    break;
	case CAIRO_PATH_LINE_TO:
	    bend_line_to (&context,
			  data[1].point.x, data[1].point.y);
	    break;
	case CAIRO_PATH_CURVE_TO:
	    g_assert_not_reached ();
	    break;
	case CAIRO_PATH_CLOSE_PATH:
	    bend_close_path (&context);
	    break;
	}
    }

    cairo_path_destroy (path);
}

/* The static MAX_PTS and MAX_POLYS limits are pretty lame, but at
 * least we exit if they are exceeded */
static void
poly_add_pt (poly_t *poly, double x, double y)
{
    poly->num_pts++;
    assert (poly->num_pts <= MAX_PTS);

    poly->pts[poly->num_pts - 1].x = x;
    poly->pts[poly->num_pts - 1].y = y;
}

static void
path_to_poly_list_move_to (poly_list_t *poly_list,
			   double       x,
			   double       y)
{
    poly_t *poly;

    poly_list->num_polys++;
    assert (poly_list->num_polys <= MAX_POLYS);

    poly = &poly_list->polys[poly_list->num_polys - 1];
    poly->num_pts = 0;

    poly_add_pt (poly, x, y);
}

static void
path_to_poly_list_line_to (poly_list_t *poly_list,
			   double       x,
			   double       y)
{
    poly_t *poly;

    if (poly_list->num_polys == 0)
	return path_to_poly_list_move_to (poly_list, x, y);

    poly = &poly_list->polys[poly_list->num_polys - 1];

    poly_add_pt (poly, x, y);
}

static void
path_to_poly_list_close_path (poly_list_t *poly_list)
{
    poly_t *poly;

    if (poly_list->num_polys == 0)
	return;

    poly = &poly_list->polys[poly_list->num_polys - 1];

    if (poly->num_pts == 0)
	return;

    poly_add_pt (poly,
		 poly->pts[0].x,
		 poly->pts[0].y);
}

static void
path_to_poly_list (cairo_t *cr, poly_list_t *poly_list)
{
    cairo_path_t *path;
    int i;

    poly_list->num_polys = 0;

    path = cairo_copy_path_flat (cr);
    for (i=0; i < path->num_data; i += path->data[i].header.length) {
	cairo_path_data_t *data = &path->data[i];

	switch (data->header.type) {
	case CAIRO_PATH_MOVE_TO:
	    path_to_poly_list_move_to (poly_list,
				       data[1].point.x, data[1].point.y);
	    break;
	case CAIRO_PATH_LINE_TO:
	    path_to_poly_list_line_to (poly_list,
				       data[1].point.x, data[1].point.y);
	    break;
	case CAIRO_PATH_CURVE_TO:
	    g_assert_not_reached ();
	    break;
	case CAIRO_PATH_CLOSE_PATH:
	    path_to_poly_list_close_path (poly_list);
	    break;
	}
    }

    cairo_path_destroy (path);
}

#define NUM_BRISTLES 2
#define PEN_WIDTH 5

static void
find_bristle_position (pt_t *b, int index, double time)
{
    double scale = PEN_WIDTH * (1.0 - 0.8 * time);

    b->x = scale * (-0.5 + (double) index / NUM_BRISTLES);
    b->y = scale * (-0.5 + (double) index / NUM_BRISTLES);
}

static double
poly_length (poly_t *poly)
{
    int i;
    double length = 0;

    for (i=0; i < poly->num_pts - 1; i++)
	length += distance (poly->pts[i], poly->pts[i+1]);

    return length;
}

static void
stroke_poly_as_brush (cairo_t *cr, poly_t *poly)
{
    int i, j;
    pt_t bristle;
    double length_total, length_so_far;
    int start, stop, step;

    cairo_save (cr);

    cairo_set_line_width (cr, 1.0);

    length_total = poly_length (poly);

    cairo_new_path (cr);

    for (i=0; i < NUM_BRISTLES; i++) {
	length_so_far = 0.0;

	if (i % 2 == 0) {
	    start = 0;
	    stop = poly->num_pts;
	    step = 1;
	} else {
	    start = poly->num_pts - 1;
	    stop = -1;
	    step = -1;
	}

	for (j=start; j != stop; j+=step) {
	    if (j != start)
		length_so_far += distance (poly->pts[j-step], poly->pts[j]);
	    find_bristle_position (&bristle, i, length_so_far / length_total);
	    
	    cairo_line_to (cr,
			   poly->pts[j].x + bristle.x,
			   poly->pts[j].y + bristle.y);
	}

    }

    cairo_close_path (cr);

    cairo_fill (cr);

    cairo_restore (cr);
}

static void
stroke_as_brush (cairo_t *cr)
{
    poly_list_t poly_list;
    int i;

    path_to_poly_list (cr, &poly_list);

    cairo_new_path (cr);

    for (i=0; i < poly_list.num_polys; i++)
	stroke_poly_as_brush (cr, &poly_list.polys[i]);
}

void
caligula_draw_sketchy_box (cairo_t *cr, int x, int y, int width, int height)
{
    cairo_rectangle (cr, x, y, width, height);
    bend_path (cr);
    stroke_as_brush (cr);
}
