/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 * Based on CGE by Julien Boulnois.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "caligula-style.h"
#include "caligula-rc-style.h"

#include <gtk/gtk.h>
#include <string.h>

static void
caligula_rc_style_init (CaligulaRcStyle *style);

static void
caligula_rc_style_class_init (CaligulaRcStyleClass *klass);

static void
caligula_rc_style_finalize (GObject *object);

static GtkStyle *
caligula_rc_style_create_style (GtkRcStyle *style);

static guint
caligula_rc_style_parse (GtkRcStyle *rc_style,
			 GtkSettings  *settings,
			 GScanner   *scanner);

static void
caligula_rc_style_merge (GtkRcStyle *dest,
			 GtkRcStyle *src);

static GtkRcStyleClass *parent_class;

GType caligula_type_rc_style = 0;

void
caligula_rc_style_register_type (GTypeModule *module)
{
    static const GTypeInfo object_info = {
	sizeof (CaligulaRcStyleClass),
	(GBaseInitFunc) NULL,
	(GBaseFinalizeFunc) NULL,
	(GClassInitFunc) caligula_rc_style_class_init,
	NULL,           /* class_finalize */
	NULL,           /* class_data */
	sizeof (CaligulaRcStyle),
	0,              /* n_preallocs */
	(GInstanceInitFunc) caligula_rc_style_init,
    };
  
    caligula_type_rc_style = g_type_module_register_type (module,
							  GTK_TYPE_RC_STYLE,
							  "CaligulaRcStyle",
							  &object_info, 0);
}

static void
caligula_rc_style_init (CaligulaRcStyle *rc_style)
{
    rc_style->drawing_style = CALIGULA_DRAWING_STYLE_DEFAULT;
    rc_style->window_top_color.pixel = FALSE;
    rc_style->window_bottom_color.pixel = FALSE;
}

static void
caligula_rc_style_class_init (CaligulaRcStyleClass *klass)
{
    GtkRcStyleClass *rc_style_class = GTK_RC_STYLE_CLASS (klass);
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    parent_class = g_type_class_peek_parent (klass);

    rc_style_class->parse = caligula_rc_style_parse;
    rc_style_class->merge = caligula_rc_style_merge;
    rc_style_class->create_style = caligula_rc_style_create_style;

    object_class->finalize = caligula_rc_style_finalize;
}

static void
caligula_rc_style_finalize (GObject *object)
{
    CaligulaRcStyle *rc_style = CALIGULA_RC_STYLE (object);

    /* We would free anything necessary in rc_style here. */

    /* For now, just keep the compiler quiet about unused variables */
    rc_style->drawing_style = rc_style->drawing_style;
  
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static GtkStyle *
caligula_rc_style_create_style (GtkRcStyle *style)
{
    return g_object_new (CALIGULA_TYPE_STYLE, NULL);
}

enum
{
    TOKEN_DRAWING_STYLE = G_TOKEN_LAST + 1,
    TOKEN_WINDOW_TOP_COLOR,
    TOKEN_WINDOW_BOTTOM_COLOR
};

static struct
{
    gchar *name;
    guint token;
}
theme_symbols[] =
{
    { "drawing-style",		TOKEN_DRAWING_STYLE },
    { "window-top-color",	TOKEN_WINDOW_TOP_COLOR },
    { "window-bottom-color",	TOKEN_WINDOW_BOTTOM_COLOR }
};

static GTokenType
caligula_rc_style_parse_drawing_style (CaligulaRcStyle	*rc_style,
				       GtkSettings	*settings,
				       GScanner		*scanner)
{
    GTokenType token;
    GTokenValue value;

    token = g_scanner_get_next_token (scanner);
    if (token != TOKEN_DRAWING_STYLE)
	return TOKEN_DRAWING_STYLE;

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
	return G_TOKEN_EQUAL_SIGN;

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_IDENTIFIER)
	return G_TOKEN_IDENTIFIER;

    value = g_scanner_cur_value (scanner);

    if (strcmp (value.v_identifier, "grassy") == 0)
	rc_style->drawing_style = CALIGULA_DRAWING_STYLE_GRASSY;
    else if (strcmp (value.v_identifier, "sketchy") == 0)
	rc_style->drawing_style = CALIGULA_DRAWING_STYLE_SKETCHY;
    else if (strcmp (value.v_identifier, "swirly") == 0)
	rc_style->drawing_style = CALIGULA_DRAWING_STYLE_SWIRLY;
    else
	g_warning ("%s: Ignoring unknown drawing-style value: %s\n",
		   __FILE__, value.v_identifier);

    return G_TOKEN_NONE;
}

static GTokenType
caligula_rc_style_parse_window_top_color (CaligulaRcStyle	*rc_style,
					  GtkSettings	        *settings,
					  GScanner		*scanner)
{
    GTokenType token;

    token = g_scanner_get_next_token (scanner);
    if (token != TOKEN_WINDOW_TOP_COLOR)
	return TOKEN_WINDOW_TOP_COLOR;

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
	return G_TOKEN_EQUAL_SIGN;

    token = gtk_rc_parse_color (scanner,
				&rc_style->window_top_color);
    if (token != G_TOKEN_NONE)
	return token;

    rc_style->window_top_color.pixel = TRUE;

    return G_TOKEN_NONE;
}

static GTokenType
caligula_rc_style_parse_window_bottom_color (CaligulaRcStyle	*rc_style,
					     GtkSettings	*settings,
					     GScanner		*scanner)
{
    GTokenType token;

    token = g_scanner_get_next_token (scanner);
    if (token != TOKEN_WINDOW_BOTTOM_COLOR)
	return TOKEN_WINDOW_BOTTOM_COLOR;

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
	return G_TOKEN_EQUAL_SIGN;

    token = gtk_rc_parse_color (scanner,
				&rc_style->window_bottom_color);
    if (token != G_TOKEN_NONE)
	return token;

    rc_style->window_bottom_color.pixel = TRUE;

    return G_TOKEN_NONE;
}

static GTokenType
caligula_rc_style_parse (GtkRcStyle  *rc_style,
			 GtkSettings *settings,
			 GScanner    *scanner)
{
  static GQuark scope_id = 0;
  CaligulaRcStyle *caligula_style = CALIGULA_RC_STYLE (rc_style);

  guint old_scope;
  GTokenType token, expected;
  gint i;

  /* Set up a new scope in this scanner. */
  if (!scope_id)
      scope_id = g_quark_from_string ("caligula_theme_engine");

  old_scope = g_scanner_set_scope (scanner, scope_id);

  /* Add symbols for this scope, (if not present already) */
  if (!g_scanner_lookup_symbol(scanner, theme_symbols[0].name)) {
      for (i = 0; i < G_N_ELEMENTS (theme_symbols); i++)
	  g_scanner_scope_add_symbol(scanner, scope_id,
				     theme_symbols[i].name,
				     GINT_TO_POINTER(theme_symbols[i].token));
  }

  /* Begin parsing top-level. Apparently a left curly brace has
   * already been consumed for use, but we're expected to consume the
   * final right closing brace. Weird. */
  while (1) {
      token = g_scanner_peek_next_token(scanner);
      if (token == G_TOKEN_RIGHT_CURLY) {
	  g_scanner_get_next_token (scanner);
	  expected = G_TOKEN_NONE;
	  break;
      }

      switch (token)
      {
      case TOKEN_DRAWING_STYLE:
	  expected = caligula_rc_style_parse_drawing_style (caligula_style, settings, scanner);
	  break;
      case TOKEN_WINDOW_TOP_COLOR:
	  expected = caligula_rc_style_parse_window_top_color (caligula_style, settings, scanner);
	  break;
      case TOKEN_WINDOW_BOTTOM_COLOR:
	  expected = caligula_rc_style_parse_window_bottom_color (caligula_style, settings, scanner);
	  break;
      default:
	  g_scanner_get_next_token (scanner);
	  expected = G_TOKEN_RIGHT_CURLY;
	  break;
      }

      if (expected != G_TOKEN_NONE)
	  break;
  }

  /* On success, restore the previous scope. (For errors, we need the
   * current scope to format a proper error message.) */
  if (expected == G_TOKEN_NONE)
      g_scanner_set_scope(scanner, old_scope);

  return expected;
}

static void
caligula_rc_style_merge (GtkRcStyle *dest,
			 GtkRcStyle *src)
{
    if (CALIGULA_IS_RC_STYLE (src)) {

	CaligulaRcStyle *caligula_dest = CALIGULA_RC_STYLE (dest);
	CaligulaRcStyle *caligula_src = CALIGULA_RC_STYLE (src);

	if (caligula_src->drawing_style != CALIGULA_DRAWING_STYLE_DEFAULT)
	    caligula_dest->drawing_style = caligula_src->drawing_style;

	if (caligula_src->window_top_color.pixel)
	    caligula_dest->window_top_color = caligula_src->window_top_color;

	if (caligula_src->window_bottom_color.pixel)
	    caligula_dest->window_bottom_color = caligula_src->window_bottom_color;
    }

    parent_class->merge (dest, src);
}
