/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 * Based on CGE by Julien Boulnois.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <stdlib.h>

#include <cairo.h>
#include <cairo-xlib.h>
#include <gdk/gdkx.h>
#include <gtk/gtkwindow.h>

#include "caligula.h"
#include "caligula-rand.h"
#include "caligula-style.h"
#include "caligula-rc-style.h"
#include "caligula-draw-sketchy.h"
#include "caligula-draw-swirly.h"
#include "caligula-draw-grassy.h"
#include "caligula-draw-shadowed.h"

static GtkStyleClass *caligula_style_parent_class;

static cairo_t *
caligula_begin_paint (GdkDrawable  *window,
		      GdkRectangle *area)
{
    cairo_t *cr;

    cr = gdk_cairo_create (window);

    if (area) {
	cairo_rectangle (cr, area->x, area->y, area->width, area->height);
	cairo_clip (cr);
	cairo_new_path (cr);
    }

    return cr;
}

static void
caligula_end_paint (cairo_t *cr)
{
    cairo_destroy (cr);
}

static void
caligula_draw_shadow (GtkStyle      *style,
		      GdkWindow     *window,
		      GtkStateType   state_type,
		      GtkShadowType  shadow_type,
		      GdkRectangle  *area,
		      GtkWidget     *widget,
		      const gchar   *detail,
		      int            x,
		      int            y,
		      int            width,
		      int            height)
{
    cairo_t *cr;

    cr = caligula_begin_paint (window, area);

    if (shadow_type == GTK_SHADOW_IN) {
	caligula_draw_shadowed_inset_box (cr, x, y, width, height,
					  style->xthickness, style->ythickness);
    } else {
	caligula_style_parent_class->draw_shadow (style, window,
						  state_type, shadow_type,
						  area, widget, detail,
						  x, y, width, height);
    }

    caligula_end_paint (cr);
}

static void
caligula_draw_box (GtkStyle      *style,
		   GdkWindow     *window,
		   GtkStateType   state_type,
		   GtkShadowType  shadow_type,
		   GdkRectangle  *area,
		   GtkWidget     *widget,
		   const gchar   *detail,
		   int            x,
		   int            y,
		   int            width,
		   int            height)
{ 
    cairo_t *cr;
    CaligulaRcStyle *rc_style = CALIGULA_RC_STYLE (style->rc_style);

    caligula_srand ((unsigned)widget);

    cr = caligula_begin_paint (window, area);

    switch (rc_style->drawing_style) {
    case CALIGULA_DRAWING_STYLE_SKETCHY:
	caligula_draw_sketchy_box (cr, x, y, width, height);
	break;
    case CALIGULA_DRAWING_STYLE_SWIRLY:
	caligula_draw_swirly_box (cr, x, y, width, height);
	break;
    case CALIGULA_DRAWING_STYLE_GRASSY:
    case CALIGULA_DRAWING_STYLE_DEFAULT:
    default:
	caligula_draw_grassy_box (cr, x, y, width, height);
	break;
    }

    caligula_end_paint (cr);
}

static void
draw_window_background (GtkStyle     *style,
			GtkWidget    *widget,
			GdkWindow    *window,
			GdkRectangle *area)
{
    CaligulaRcStyle *rc_style = CALIGULA_RC_STYLE (style->rc_style);
    cairo_t *cr;
    GdkColor top_color;
    GdkColor bottom_color;

    /* GtkWindow paints out of realize() to allow the
     * theme to set the window background pixmap
     */
    if (!GTK_WIDGET_MAPPED (widget))
	return;

    if (!rc_style->window_top_color.pixel)
	top_color = style->bg[widget->state];
    else
      top_color = rc_style->window_top_color;

    if (!rc_style->window_bottom_color.pixel)
	bottom_color = style->bg[widget->state];
    else
	bottom_color = rc_style->window_bottom_color;

    cr = caligula_begin_paint (window, area);

    if (top_color.red == bottom_color.red &&
	top_color.green == bottom_color.green &&
	top_color.blue == bottom_color.blue) {

	cairo_set_source_rgb (cr,
			      top_color.red / 65535.,
			      top_color.green / 65535.,
			      top_color.blue / 65535.);
	
    } else {
	cairo_pattern_t *pattern;
	int height;
	
	gdk_drawable_get_size (window, 0, &height);

	pattern = cairo_pattern_create_linear (0, 0, 0, height);
	if (pattern) {
	  cairo_pattern_add_color_stop_rgb (pattern, 0.0,
					    top_color.red / 65535.,
					    top_color.green / 65535.,
					    top_color.blue / 65535.);
	    cairo_pattern_add_color_stop_rgb (pattern, 1.0,
					      bottom_color.red / 65535.,
					      bottom_color.green / 65535.,
					      bottom_color.blue / 65535.);
	    cairo_set_source (cr, pattern);
	    cairo_pattern_destroy (pattern);
	}
    }

    cairo_paint (cr);
    
    caligula_end_paint (cr);
}

static void
caligula_draw_flat_box   (GtkStyle        *style,
			  GdkWindow       *window,
			  GtkStateType     state_type,
			  GtkShadowType    shadow_type,
			  GdkRectangle    *area,
			  GtkWidget       *widget,
			  const gchar     *detail,
			  int              x,
			  int              y,
			  int              width,
			  int              height)
{
    if (widget && GTK_IS_WINDOW (widget)) {
	draw_window_background (style, widget, window, area);
    } else {
	caligula_style_parent_class->draw_flat_box (style, window,
						    state_type, shadow_type,
						    area, widget, detail,
						    x, y, width, height);
    }
}

static void
caligula_style_init (CaligulaStyle *style)
{
}

static void
caligula_style_class_init (CaligulaStyleClass *klass)
{
    GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);

    style_class->draw_shadow = caligula_draw_shadow;
    style_class->draw_box = caligula_draw_box;
    style_class->draw_flat_box = caligula_draw_flat_box;
    
    caligula_style_parent_class = g_type_class_peek_parent (klass);
}

GType caligula_type_style = 0;

void
caligula_style_register_type (GTypeModule *module)
{
    static const GTypeInfo object_info = {
	sizeof (CaligulaStyleClass),
	(GBaseInitFunc) NULL,
	(GBaseFinalizeFunc) NULL,
	(GClassInitFunc) caligula_style_class_init,
	NULL,           /* class_finalize */
	NULL,           /* class_data */
	sizeof (CaligulaStyle),
	0,              /* n_preallocs */
	(GInstanceInitFunc) caligula_style_init,
  };
  
  caligula_type_style = g_type_module_register_type (module,
						     GTK_TYPE_STYLE,
						     "CaligulaStyle",
						     &object_info, 0);
}
