/* CGE - cairo-gtk-engine

   Copyright © 2003 Julien Boulnois

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with this program; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA
*/

#include "cge.h"

static void
cge_rc_style_init (CgeRCStyle * style);

static void
cge_rc_style_class_init (CgeRCStyleClass * klass);

static guint
cge_rc_style_parse (GtkRcStyle *rc_style, 
		    GtkSettings *settings,
		    GScanner *scanner);

static void
cge_rc_style_merge (GtkRcStyle * dest, GtkRcStyle * src);

static GtkStyle *
cge_rc_style_create_style (GtkRcStyle * rc_style);

static GtkRcStyleClass * parent_class;

GType cge_type_rc_style = 0;

enum
{
    TOKEN_STYLE= G_TOKEN_LAST + 1,
    TOKEN_DETAIL,
    TOKEN_FILL,
    TOKEN_LINE,
    TOKEN_MODE,
    TOKEN_ARROW,

    TOKEN_ROUND_WIDTH,
    TOKEN_LINE_WIDTH,

    TOKEN_GRADIENT_DIRECTION,
    TOKEN_COLOR,
    TOKEN_ALPHA,
};
  
static struct
{
    gchar       *name;
    guint        token;
}
theme_symbols[] =
{
    { "style",               TOKEN_STYLE },
    { "detail",              TOKEN_DETAIL },
    { "fill",                TOKEN_FILL },
    { "line",                TOKEN_LINE },
    { "mode",                TOKEN_MODE },
    { "arrow",               TOKEN_ARROW },
    { "round_width",         TOKEN_ROUND_WIDTH },
    { "line_width",          TOKEN_LINE_WIDTH },
    { "gradient_direction",  TOKEN_GRADIENT_DIRECTION },
    { "color",               TOKEN_COLOR },
    { "alpha",               TOKEN_ALPHA },
};

static guint n_theme_symbols = sizeof(theme_symbols) / sizeof(theme_symbols[0]);

static gboolean 
TranslateFillStyleMode (gchar * str, gint *retval)
{
#define is_enum(XX) (g_ascii_strncasecmp(str, XX, strlen(XX))==0)

    if (is_enum("solid"))
	*retval = CGE_FILL_SOLID;
    else if (is_enum("gradient"))
	*retval = CGE_FILL_GRADIENT;
    else if (is_enum("file"))
	*retval = CGE_FILL_FILE;
    else
	return FALSE; 

    return TRUE;
}

static gboolean 
TranslateLineStyleMode (gchar * str, gint *retval)
{
#define is_enum(XX)  (g_ascii_strncasecmp(str, XX, strlen(XX))==0)
    if (is_enum("solid"))
	*retval = CGE_LINE_SOLID;
    else if (is_enum("none"))
	*retval = CGE_LINE_NONE;
    else if (is_enum("dashed"))
	*retval = CGE_LINE_DASHED;
    else if (is_enum("bevel"))
	*retval = CGE_LINE_BEVEL;
    else if (is_enum("file"))
	*retval = CGE_LINE_FILE;
    else
	return FALSE; 
    return TRUE;
}

static gboolean 
TranslateGradientDirection (gchar * str, gint *retval)
{
#define is_enum(XX)  (g_ascii_strncasecmp(str, XX, strlen(XX))==0)
    if (is_enum("vertical"))
	*retval = CGE_DIRECTION_VERTICAL;
    else if (is_enum("horizontal"))
	*retval = CGE_DIRECTION_HORIZONTAL;
    else
	return FALSE; 
    return TRUE;
}

static gboolean 
TranslateArrowStyle (gchar * str, gint *retval)
{
#define is_enum(XX)  (g_ascii_strncasecmp(str, XX, strlen(XX))==0)
    if (is_enum("normal"))
	*retval = CGE_ARROW_NORMAL;
    else if (is_enum("simple"))
	*retval = CGE_ARROW_SIMPLE;
    else if (is_enum("file"))
	*retval = CGE_ARROW_FILE;
    else
	return FALSE; 
    return TRUE;
}

static guint 
theme_parse_custom_enum (GScanner *scanner, 
                         GTokenType wanted_token, 
			 CgeTranslateEnumFunc translate_enum,
			 guint return_default,
			 int *retval)
{
    guint token;
   
    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
    {
	return G_TOKEN_EQUAL_SIGN;
    }

    token = g_scanner_get_next_token (scanner);

    if (token != G_TOKEN_IDENTIFIER)
	*retval = return_default;
    else 
	if (!translate_enum(scanner->value.v_identifier, (gint *) retval))
	    *retval = return_default; 

    return G_TOKEN_NONE;
}

static guint 
theme_parse_double(GScanner *scanner, 
		   GTokenType wanted_token, 
		   gdouble return_default, 
		   gdouble *retval, 
		   gdouble lower_limit, 
		   gdouble upper_limit)
{
    guint token;

    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
    {
	return G_TOKEN_EQUAL_SIGN;
    }

    token = g_scanner_get_next_token (scanner);

    if (token != G_TOKEN_FLOAT)
	*retval = return_default;
    else
	*retval = scanner->value.v_float;
    
    if (*retval < lower_limit) *retval = lower_limit; 
    if ((*retval > upper_limit) && (upper_limit > lower_limit)) *retval = upper_limit; 
  
    return G_TOKEN_NONE;
}

static guint 
theme_parse_char(GScanner *scanner, 
                 GTokenType wanted_token, 
		 gchar **retval)

{
    guint token;

    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_EQUAL_SIGN)
    {
	return G_TOKEN_EQUAL_SIGN;
    }

    token = g_scanner_get_next_token (scanner);

    /* printf("detail : %s\n\n",scanner->value.v_string); */
    *retval = g_strdup(scanner->value.v_string);
  
    return G_TOKEN_NONE;
}

static guint
theme_parse_fill (GtkSettings	 *settings,
		  GScanner	 *scanner,
		  GTokenType	 wanted_token,
		  cge_fill_style *retval)
{
    guint token;
    GdkColor color[5];
    gint colors_count[5]={0,0,0,0,0};
    gint i;

    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_LEFT_CURLY)
	return G_TOKEN_LEFT_CURLY;

    token = g_scanner_peek_next_token (scanner);
    while (token != G_TOKEN_RIGHT_CURLY) {
	switch (token) {
	case TOKEN_MODE:
	    token = theme_parse_custom_enum(scanner, TOKEN_MODE, TranslateFillStyleMode, CGE_FILL_SOLID,  (int*)&retval->fill_mode);
	    break;
	case TOKEN_GRADIENT_DIRECTION:
	    token = theme_parse_custom_enum(scanner, TOKEN_GRADIENT_DIRECTION, TranslateGradientDirection, CGE_DIRECTION_VERTICAL, (int*)&retval->gradient_direction);
	case TOKEN_COLOR:
	{
	    GtkStateType state;
	    
	    token = g_scanner_get_next_token (scanner);
	    if (token != TOKEN_COLOR)
		return TOKEN_COLOR;
	    
	    token = gtk_rc_parse_state (scanner, &state);
	    if (token != G_TOKEN_NONE)
		return token;
	    
	    token = g_scanner_get_next_token (scanner);
	    if (token != G_TOKEN_EQUAL_SIGN)
		return G_TOKEN_EQUAL_SIGN;
	    
	    token = gtk_rc_parse_color (scanner, &color[state]);
	
	    retval->gradient_colors[state][colors_count[state]]=color[state];
	    colors_count[state]++;	
	}
	break;
	case TOKEN_ROUND_WIDTH:      
	    token = theme_parse_double (scanner, TOKEN_ROUND_WIDTH, 4.0, &retval->round_width, 0.0, 10.0);
	    break;
	case TOKEN_ALPHA:      
	    token = theme_parse_double (scanner, TOKEN_ALPHA, 1.0, &retval->alpha, 0.0, 1.0);
	    break;
	default:
	    g_scanner_get_next_token (scanner);
	    token = G_TOKEN_RIGHT_CURLY;
	    break;
	}

	token = g_scanner_peek_next_token (scanner);
    }

    g_scanner_get_next_token (scanner);
    token = G_TOKEN_NONE;

    for(i=0;i<5;i++)
    {
	if (colors_count[i]>0) 
	{
	    retval->gradient_colors_count[i]=colors_count[i];
	    retval->gradient_colors_default[i]=FALSE;
	    
	}
    }

    return token;
}

static guint
theme_parse_line (GtkSettings	 *settings,
		  GScanner	 *scanner,
		  GTokenType	 wanted_token,
		  cge_line_style *retval)
{
    guint token;

    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_LEFT_CURLY)
	return G_TOKEN_LEFT_CURLY;

    token = g_scanner_peek_next_token (scanner);
    while (token != G_TOKEN_RIGHT_CURLY) {
	switch (token) {
	case TOKEN_MODE:
	    token = theme_parse_custom_enum(scanner, TOKEN_MODE, TranslateLineStyleMode, CGE_LINE_SOLID,  (int*)&retval->line_mode);
	    break;
	case TOKEN_LINE_WIDTH:
	    token = theme_parse_double (scanner, TOKEN_LINE_WIDTH, 1.0, &retval->line_width, 0.0, 10.0);
	    break;
	case TOKEN_COLOR:
	{
	    GtkStateType state;
	    
	    token = g_scanner_get_next_token (scanner);
	    if (token != TOKEN_COLOR)
		return TOKEN_COLOR;
	    
	    token = gtk_rc_parse_state (scanner, &state);
	    if (token != G_TOKEN_NONE)
		return token;
	    
	    token = g_scanner_get_next_token (scanner);
	    if (token != G_TOKEN_EQUAL_SIGN)
		return G_TOKEN_EQUAL_SIGN;
	    
	    token = gtk_rc_parse_color (scanner, &retval->color[state]);
	    retval->color_default[state]=FALSE;
	}
	break;
	case TOKEN_ROUND_WIDTH:      
	    token = theme_parse_double (scanner, TOKEN_ROUND_WIDTH, 4.0, &retval->round_width, 0.0, 10.0);
	    break;
	case TOKEN_ALPHA:      
	    token = theme_parse_double (scanner, TOKEN_ALPHA, 1.0, &retval->alpha, 0.0, 1.0);
	    break;      
	default:
	    g_scanner_get_next_token (scanner);
	    token = G_TOKEN_RIGHT_CURLY;
	    break;
	}

	token = g_scanner_peek_next_token (scanner);
    }

    g_scanner_get_next_token (scanner);
    token = G_TOKEN_NONE;

    return token;
}


static guint
theme_parse_style (GtkSettings	*settings,
		   GScanner	*scanner,
		   GTokenType	wanted_token,
		   cge_style	*retval)
{
    guint token;

    token = g_scanner_get_next_token (scanner);
    if (token != wanted_token)
    {
	return wanted_token;
    }

    token = g_scanner_get_next_token (scanner);
    if (token != G_TOKEN_LEFT_CURLY)
	return G_TOKEN_LEFT_CURLY;

    token = g_scanner_peek_next_token (scanner);
    while (token != G_TOKEN_RIGHT_CURLY) {
	switch (token) {
	case TOKEN_DETAIL:
	    token = theme_parse_char(scanner, TOKEN_DETAIL, &retval->detail);
	    /* printf("style %s found\n",retval->detail); */
	    break;
	case TOKEN_FILL :
	    token = theme_parse_fill (settings, scanner, TOKEN_FILL, &retval->fill_style);
	    break;
	case TOKEN_LINE:
	    token = theme_parse_line (settings, scanner, TOKEN_LINE, &retval->line_style);      
	    break;
	default:
	    g_scanner_get_next_token (scanner);
	    token = G_TOKEN_RIGHT_CURLY;
	    break;
	}

	token = g_scanner_peek_next_token (scanner);
    }

    g_scanner_get_next_token (scanner);
    token = G_TOKEN_NONE;

    return token;
}


void
cge_rc_style_register_type (GTypeModule * module)
{
  static const GTypeInfo object_info = {
      sizeof (CgeRCStyleClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) cge_rc_style_class_init,
      NULL,
      NULL,
      sizeof (CgeRCStyle),
      0,
      (GInstanceInitFunc) cge_rc_style_init,
  };

  cge_type_rc_style = g_type_module_register_type (module,
						   GTK_TYPE_RC_STYLE,
						   "CgeRCStyle",
						   &object_info, 0);
}

static void
cge_rc_init_fill_style (cge_fill_style* fill_style)
{
    gint i;

    fill_style->fill_mode=CGE_FILL_SOLID;
    fill_style->round_width=4;
    fill_style->alpha=1;
    fill_style->gradient_direction=CGE_DIRECTION_VERTICAL;

    for(i=0;i<5;i++)
    { 
	fill_style->gradient_colors_count[i]=2;
	fill_style->gradient_colors_default[i]=TRUE;
    }
}

static void
cge_rc_init_line_style (cge_line_style* line_style)
{
    gint i;

    for(i=0;i<5;i++)
    {
	line_style->color_default[i]=TRUE;
    }

    line_style->line_mode=CGE_LINE_SOLID;
    line_style->line_width=1;
    line_style->round_width=4;
    line_style->alpha=1;
}

static void
cge_rc_style_init (CgeRCStyle * style)
{
    gint i;

    for(i=0;i<MAX_STYLES;i++)
    {
	style->styles[i].detail=NULL;
	cge_rc_init_fill_style(&style->styles[i].fill_style);
	cge_rc_init_line_style(&style->styles[i].line_style);
    }

    style->arrow_style=CGE_ARROW_NORMAL;
}

static void
cge_rc_style_class_init (CgeRCStyleClass * klass)
{
    GtkRcStyleClass * rc_style_class = GTK_RC_STYLE_CLASS (klass);

    parent_class = g_type_class_peek_parent (klass);

    rc_style_class->parse = cge_rc_style_parse;
    rc_style_class->merge = cge_rc_style_merge;
    rc_style_class->create_style = cge_rc_style_create_style;
}


static guint
cge_rc_style_parse (GtkRcStyle  *rc_style, 
		    GtkSettings *settings,
		    GScanner	*scanner)
{

    static GQuark scope_id = 0;
    CgeRCStyle *cge_style = NEW_THEME_DATA(rc_style);
  
    guint old_scope;
    guint token;
    guint i;
    gint styles_count=1;

    /* Set up a new scope in this scanner */
    if (!scope_id)
    {
	scope_id = g_quark_from_string ("cge_theme_engine");
    }

    /* If we bail out due to errors, we don't reset the scope, so the
     * error messaging code can make sense of our tokens. */
    old_scope = g_scanner_set_scope (scanner, scope_id);

    /* Check if we already added our symbols to this scope */
    if (!g_scanner_lookup_symbol (scanner, theme_symbols[0].name))
    {
	for (i = 0; i < n_theme_symbols; i++)
	{
	    g_scanner_scope_add_symbol (scanner, scope_id, theme_symbols[i].name, 
					GINT_TO_POINTER (theme_symbols[i].token));
	}
    }

    /* Ready to go, now parse the top level */
    token = g_scanner_peek_next_token (scanner);
    while (token != G_TOKEN_RIGHT_CURLY)
    {
	switch (token)
	{
	case TOKEN_STYLE :
	    token = theme_parse_style (settings, scanner, TOKEN_STYLE, &cge_style->styles[styles_count]);
	    styles_count++;
	    break;
	case TOKEN_FILL :
	    token = theme_parse_fill (settings, scanner, TOKEN_FILL, &cge_style->styles[0].fill_style);
	    break;
	case TOKEN_LINE:
	    token = theme_parse_line (settings, scanner, TOKEN_LINE, &cge_style->styles[0].line_style);      
	    break;
	case TOKEN_ARROW:
	    token = theme_parse_custom_enum(scanner, TOKEN_ARROW, TranslateArrowStyle, CGE_ARROW_NORMAL,  (int*)&cge_style->arrow_style);
	    break;
	default:
	    g_scanner_get_next_token (scanner);
	    token = G_TOKEN_RIGHT_CURLY;
	    break;
	}

	if (token != G_TOKEN_NONE)
	{
	    g_free (cge_style);
	    return token;
	}

	token = g_scanner_peek_next_token (scanner);
    }

    g_scanner_get_next_token(scanner);

    cge_style->styles[0].detail="all";

    SET_THEME_DATA(rc_style, cge_style);

    g_scanner_set_scope (scanner, old_scope);
  
    return G_TOKEN_NONE;
}

static void
cge_rc_style_merge (GtkRcStyle * dest, GtkRcStyle * src)
{
    gint i;

    parent_class->merge (dest, src);

    if (CGE_IS_RC_STYLE (src)) 
    {     
	CgeRCStyle *src_data = CGE_RC_DATA (src);
	CgeRCStyle *dest_data = CGE_RC_DATA (dest);

	for(i=0;i<MAX_STYLES;i++)
	{
	    dest_data->styles[i]=src_data->styles[i];
	}

	dest_data->arrow_style=src_data->arrow_style;
	SET_THEME_DATA(dest, dest_data);
    }
}

static GtkStyle *
cge_rc_style_create_style (GtkRcStyle * rc_style)
{
    void * ptr;

    ptr = g_object_new (CGE_TYPE_STYLE, NULL);
    return (GtkStyle *) ptr;
}
